import { EditorState, Extension, Facet, StateEffect, TransactionSpec } from "@codemirror/state";
import { EditorView } from "@codemirror/view";
import { CommandPanelMessage } from "./panels";
type ExternalCommand = "file_picker" | "buffer_picker" | ":buffer-next" | ":buffer-close" | ":buffer-previous";
type ExternalCommandHandler = () => CommandPanelMessage | void;
type ExternalCommandsDefinition = Partial<Record<ExternalCommand, ExternalCommandHandler>> & {
    global_search?(input: string): void | CommandPanelMessage;
};
/**
 * A facet that allows to define external commands.
 */
declare const externalCommandsFacet: Facet<ExternalCommandsDefinition, ExternalCommandsDefinition>;
export { externalCommandsFacet as externalCommands };
/**
 * Creates a snapshot of the extension state suitable to initialize
 * the extension later (see `init` and `globalInit`). Snapshots are JSON-serializable.
 *
 * If `global` is true, the snapshot only contains global state. This way
 * it is slimmer, but it is only valid for `globalInit`.
 */
export declare function snapshot(state: EditorState, global?: boolean): Object;
/**
 * Generates a list of transactions that can be dispatched to
 * another editor to ensure that its global state is synchronized
 * with `state`.
 */
export declare function globalStateSync(state: EditorState): TransactionSpec[];
/**
 * A facet to define typable commands. No effort is made to prevent overrides,
 * collisions, etc.
 */
export declare const commands: Facet<TypableCommand[], TypableCommand[]>;
/**
 * An effect to reset the mode of an editor.
 */
export declare const resetMode: StateEffect<any>;
/**
 * A command that can be typed in command mode `:`.
 */
export interface TypableCommand {
    name: string;
    aliases?: string[];
    help: string;
    /**
     * The handler for the command. The return type can specify a message,
     * and qualify it as an error if desired.
     */
    handler(view: EditorView, args: any[]): CommandPanelMessage | void;
}
export interface Options {
    config?: Config;
    /**
     * If provided, sets the extension initial state from a previous state, or a snapshot
     * created by `snapshot()`.
     */
    init?: EditorState | Object;
    /**
     * Like `init`, but it will only restore global state that should be shared between different "tabs".
     * For instance, registers are global state, while undo/redo history is not.
     */
    globalInit?: EditorState | Object;
}
/**
 * Editor configuration.
 * The names follow Helix's options' naming.
 */
export interface Config {
    "editor.cursor-shape.insert"?: "block" | "bar";
}
/**
 * The main helix extension.
 *
 * It provides Helix-like keybindings, plus two panels to emulate the statusline and the commandline.
 */
export declare function helix(options?: Options): Extension;
