// src/lib.ts
import {
  cursorCharLeft,
  cursorCharRight,
  cursorDocEnd,
  cursorDocStart,
  deleteCharBackward,
  deleteCharForward,
  indentLess,
  indentMore,
  insertNewlineAndIndent,
  selectAll,
  selectDocEnd,
  selectDocStart,
  selectGroupLeft,
  selectGroupRight,
  selectParentSyntax,
  toggleComment
} from "@codemirror/commands";
import {
  EditorSelection as EditorSelection4,
  EditorState as EditorState3,
  Facet,
  Transaction as Transaction2
} from "@codemirror/state";
import {
  Decoration,
  EditorView as EditorView2,
  ViewPlugin,
  WidgetType,
  drawSelection,
  getPanel,
  keymap,
  showPanel
} from "@codemirror/view";
import { SearchQuery } from "@codemirror/search";

// src/state.ts
import {
  StateEffect,
  StateField
} from "@codemirror/state";
var modeEffect = StateEffect.define();
var modeField = StateField.define({
  create() {
    return { type: 0 /* Normal */, minor: 2 /* Normal */ };
  },
  update(mode, tr) {
    for (const effect of tr.effects) {
      if (effect.is(modeEffect)) {
        mode = effect.value;
      }
    }
    return mode;
  }
});
function sameMode(mode, otherMode) {
  return mode.type === otherMode.type && mode.minor === otherMode.minor;
}
function sameModeState(mode, otherMode) {
  return sameMode(mode, otherMode) && mode.count === otherMode.count && mode.expecting === otherMode.expecting;
}
function modeStatus(mode) {
  let result = "";
  if (mode.type === 1 /* Insert */) {
    return result;
  }
  if (mode.count) {
    result += mode.count;
  }
  result += minorModeStr(mode.minor);
  if (mode.expecting) {
    result += mode.expecting.minor;
  }
  return result;
}
function minorModeStr(minor) {
  switch (minor) {
    case 2 /* Normal */:
      return "";
    case 3 /* Goto */:
      return "g";
    case 5 /* Match */:
      return "m";
    case 6 /* Space */:
      return "<space>";
    case 7 /* LeftBracket */:
      return "[";
    case 8 /* RightBracket */:
      return "]";
    default: {
      if (true) {
        throw new Error("Unexpected mode");
      }
    }
  }
}
var yankEffect = StateEffect.define();
var registersField = StateField.define({
  create() {
    return {};
  },
  update(registers, tr) {
    for (const effect of tr.effects) {
      if (effect.is(yankEffect)) {
        if (!Array.isArray(effect.value)) {
          return effect.value.reset;
        }
        const [reg, value] = effect.value;
        if (value.length === 0) {
          const { [reg]: _reg, ...rest } = registers;
          registers = rest;
        } else {
          registers = { ...registers, [reg]: value };
        }
      }
    }
    return registers;
  }
});
var historyEffect = StateEffect.define();
var historyField = StateField.define({
  create() {
    return { checkpoints: [], cursor: -1, pending: null, head: null };
  },
  update(value, tr) {
    for (const effect of tr.effects) {
      if (effect.is(historyEffect)) {
        const effectValue = effect.value;
        switch (effectValue.type) {
          case "move": {
            const cursor = value.cursor + effectValue.offset;
            value = {
              checkpoints: value.checkpoints,
              pending: null,
              cursor,
              head: value.head ?? effectValue.head ?? null
            };
            break;
          }
          case "add": {
            if (true) {
              if (value.pending && effectValue.temp) {
                throw new Error("Unexpected temp");
              }
            }
            const checkpoints = value.cursor === -1 ? value.checkpoints : value.checkpoints.slice(value.cursor + 1);
            value = effectValue.temp ? {
              checkpoints: value.checkpoints,
              cursor: value.cursor,
              pending: effectValue.state,
              head: value.head
            } : {
              checkpoints: [effectValue.state, ...checkpoints],
              cursor: -1,
              pending: null,
              head: null
            };
            break;
          }
          case "commit": {
            if (true) {
              if (!value.pending) {
                throw new Error("unexpected commit");
              }
            }
            if (effectValue.state.doc.eq(value.pending.doc)) {
              value = {
                ...value,
                pending: null
              };
              break;
            }
            const checkpoints = value.cursor === -1 ? value.checkpoints : value.checkpoints.slice(value.cursor + 1);
            value = {
              checkpoints: [value.pending, ...checkpoints],
              cursor: -1,
              pending: null,
              head: null
            };
            break;
          }
        }
      }
    }
    return value;
  }
});
var syntaxHistoryEffect = StateEffect.define();
var syntaxHistoryField = StateField.define({
  create() {
    return { selections: [], cursor: 0, head: null, frozen: false };
  },
  update(value, tr) {
    for (const effect of tr.effects) {
      if (effect.is(syntaxHistoryEffect)) {
        switch (effect.value.type) {
          case "freeze": {
            value = { ...value, frozen: effect.value.frozen ?? true };
            break;
          }
          case "add": {
            value = {
              selections: [...value.selections, effect.value.prev],
              cursor: value.cursor + 1,
              head: effect.value.next,
              frozen: false
            };
            break;
          }
          case "move": {
            value = {
              ...value,
              cursor: value.cursor + effect.value.offset
            };
            break;
          }
          case "reset": {
            if (!value.frozen) {
              value = {
                frozen: false,
                selections: [],
                cursor: 0,
                head: null
              };
            }
          }
        }
      }
    }
    return value;
  }
});
function expandSyntaxHistory(state, expand, done) {
  const history = state.field(syntaxHistoryField);
  if (history.cursor < history.selections.length) {
    done({
      selection: history.cursor === history.selections.length - 1 ? history.head : history.selections[history.cursor + 1],
      effects: syntaxHistoryEffect.of({ type: "move", offset: 1 })
    });
    return;
  }
  expand(
    {
      effects: syntaxHistoryEffect.of({ type: "freeze" })
    },
    (tr) => {
      done(
        tr.selection?.eq(state.selection) ? {
          effects: syntaxHistoryEffect.of({
            type: "freeze",
            frozen: false
          })
        } : {
          selection: tr.selection,
          changes: tr.changes,
          scrollIntoView: tr.scrollIntoView,
          effects: [
            ...tr.effects,
            syntaxHistoryEffect.of({
              type: "add",
              prev: state.selection,
              next: tr.newSelection
            })
          ]
        }
      );
    }
  );
}
function undoSyntaxHistory(state) {
  const history = state.field(syntaxHistoryField);
  if (history.cursor > 0) {
    return {
      selection: history.selections[history.cursor - 1],
      effects: syntaxHistoryEffect.of({ type: "move", offset: -1 })
    };
  }
}

// src/panels.ts
import { EditorView } from "@codemirror/view";
import { EditorSelection as EditorSelection2 } from "@codemirror/state";
var panelStyles = EditorView.theme({
  ".cm-hx-status-panel": {
    display: "flex",
    "justify-content": "space-between",
    "font-family": "monospace"
  },
  ".cm-hx-command-panel": {
    display: "flex",
    justifyContent: "space-between",
    fontFamily: "monospace",
    minHeight: "18px"
  },
  ".cm-hx-command-input": {
    fontFamily: "monospace",
    fontSize: "inherit",
    border: "none",
    outline: "none",
    padding: "0",
    margin: "0",
    background: "inherit"
  },
  ".cm-hx-command-popup": {
    position: "fixed",
    background: "#ccc"
  },
  ".cm-hx-command-help": {
    border: "1px solid #777",
    background: "#ddd",
    padding: "2px",
    whiteSpace: "preserve"
  }
});
var CommandPanel = class {
  constructor(view, commandFacet, startSearch2) {
    this.view = view;
    this.commandFacet = commandFacet;
    this.startSearch = startSearch2;
    this.dom = $el("div");
    this.minorCommand = $el("span");
    this.inputContainer = $el("span");
    this.commandPopup = $el("div");
    this.dom.append(this.inputContainer);
    this.dom.append(this.minorCommand);
    this.dom.append(this.commandPopup);
    this.dom.classList.add("cm-hx-command-panel");
    $style(this.inputContainer, { visibility: "hidden" });
    this.label = $el("span");
    this.inputContainer.append(this.label);
    this.commandPopup.classList.add("cm-hx-command-popup");
    this.help = $el("div");
    this.autocomplete = $el("div");
    this.help.hidden = true;
    this.help.classList.add("cm-hx-command-help");
    this.autocomplete.classList.add("cm-hx-command-autocomplete");
    this.commandPopup.append(this.help);
    this.commandPopup.append(this.autocomplete);
    $style(this.minorCommand, { minWidth: "8em", textAlign: "center " });
  }
  dom;
  minorCommand;
  inputContainer;
  label;
  message = false;
  commandPopup;
  autocomplete;
  help;
  popupRequest;
  showSearchInput(global = false) {
    const input = this.searchInput(global);
    this.showInput(input, global ? "global-search:" : "search:");
  }
  showCommandInput() {
    const input = this.commandInput();
    this.showInput(input, ":");
  }
  showMinor(command) {
    this.minorCommand.textContent = modeStatus(command);
  }
  showInput(input, label) {
    this.label.textContent = label;
    $style(this.label, { color: "" });
    this.inputContainer.append(input);
    $style(this.inputContainer, { visibility: "" });
    input.focus();
  }
  createInput({
    onInput,
    onClose,
    placeholder
  }) {
    const input = $el("input");
    if (placeholder) {
      input.placeholder = placeholder;
    }
    input.classList.add("cm-hx-command-input");
    input.type = "text";
    let open = true;
    input.addEventListener("blur", () => {
      if (open) {
        onClose(false, input.value);
      }
    });
    input.addEventListener("input", () => {
      onInput(input.value);
    });
    input.addEventListener("keydown", (event) => {
      if (event.isComposing) {
        return;
      }
      const isEnter = event.key === "Enter";
      if (isEnter || event.key === "Escape") {
        open = false;
        onClose(isEnter, input.value);
      }
    });
    return input;
  }
  commandInput() {
    const { view } = this;
    const initialSelection = view.state.selection;
    const initialScroll = view.scrollSnapshot();
    const isNumber = (cmd) => /^\d+$/.test(cmd);
    return this.createInput({
      onClose: (commit, value) => {
        this.hidePopup();
        const [cmd, ...args] = value.split(/ +/);
        if (commit && cmd) {
          const commands2 = view.state.facet(this.commandFacet);
          const command = commands2.find(
            (command2) => command2.name === cmd || command2.aliases?.some((alias) => alias === cmd)
          );
          const result = command ? command.handler(view, args) : {
            message: `no such command: '${cmd}'`,
            error: true
          };
          if (result) {
            this.showMessageAndCloseInput(result);
            return;
          }
        } else if (!commit && isNumber(cmd)) {
          view.dispatch({
            selection: initialSelection
          });
          setTimeout(() => {
            view.dispatch({
              effects: initialScroll
            });
          });
        }
        this.closeInput();
      },
      onInput: (value) => {
        const cmd = value.split(/ +/).at(0);
        if (!cmd) {
          this.hidePopup();
          return;
        }
        if (isNumber(cmd)) {
          const lineNo = Number(cmd);
          if (lineNo >= 1 && lineNo <= view.state.doc.lines) {
            const line = view.state.doc.line(lineNo);
            view.dispatch({
              selection: EditorSelection2.cursor(line.from),
              effects: EditorView.scrollIntoView(line.from, { y: "center" })
            });
          }
          return;
        }
        const commands2 = view.state.facet(this.commandFacet);
        const options = commands2.filter(
          (command) => command.name.startsWith(cmd) || command.aliases?.some((alias) => alias.startsWith(cmd))
        );
        if (options.length === 0) {
          this.hidePopup();
          return;
        }
        const match = options.find(
          (command) => command.name === cmd || command.aliases?.some((alias) => alias === cmd)
        );
        this.showPopup(options, match);
      }
    });
  }
  showError(message) {
    this.showMessage({ message, error: true });
  }
  showMessage(messageOrResult) {
    if (messageOrResult == null) {
      return;
    }
    const [message, error] = typeof messageOrResult === "string" ? [messageOrResult, false] : [messageOrResult.message, messageOrResult.error];
    $style(this.inputContainer, { visibility: "" });
    this.message = true;
    $style(this.label, { color: error ? "red" : "" });
    this.label.textContent = message;
  }
  showMessageAndCloseInput(result) {
    this.showMessage(result);
    this.closeInput(!result);
  }
  hasMessage() {
    return this.message;
  }
  clearMessage() {
    if (this.message) {
      this.message = false;
      this.label.textContent = "";
      $style(this.inputContainer, { visibility: "hidden" });
    }
  }
  showPopup(commands2, match) {
    this.commandPopup.hidden = false;
    this.help.hidden = !match;
    if (match) {
      this.help.textContent = `${match.help}`;
      if (match.aliases && match.aliases.length > 0) {
        this.help.textContent += `
Aliases: ${match.aliases.join(",")}`;
      }
    } else {
      this.help.textContent = "";
    }
    while (commands2.length > this.autocomplete.childNodes.length) {
      const entry = $el("span");
      $style(entry, { marginRight: "1em" });
      this.autocomplete.append(entry);
    }
    for (const [i, child] of this.autocomplete.childNodes.entries()) {
      const command = commands2[i];
      if (command) {
        child.textContent = command.name;
      } else {
        break;
      }
    }
    while (this.autocomplete.childNodes.length > commands2.length) {
      this.autocomplete.lastChild?.remove();
    }
    if (this.popupRequest == null) {
      this.popupRequest = requestAnimationFrame(() => this.positionPopup());
    }
  }
  hidePopup() {
    this.commandPopup.hidden = true;
  }
  positionPopup() {
    this.popupRequest = void 0;
    if (this.commandPopup.hidden) {
      return;
    }
    const box = this.inputContainer.getBoundingClientRect();
    $style(this.commandPopup, {
      bottom: `${window.innerHeight - box.top}px`,
      left: `${box.left}px`
    });
  }
  searchInput(global) {
    const search = this.startSearch(global);
    return this.createInput({
      placeholder: search.init,
      onClose: (commit) => {
        this.showMessageAndCloseInput(search.onClose(commit));
      },
      onInput: (value) => {
        search.onInput(value);
      }
    });
  }
  closeInput(hide = true) {
    this.inputContainer.removeChild(this.inputContainer.lastChild);
    if (hide) {
      $style(this.inputContainer, { visibility: "hidden" });
    }
    requestAnimationFrame(() => {
      this.view.focus();
    });
  }
};
function statusPanel(view) {
  const dom = $el("div");
  dom.classList.add("cm-hx-status-panel");
  const mode = $el("span");
  mode.textContent = "NOR";
  dom.append(mode);
  const pos = $el("span");
  dom.append(pos);
  function setLineCol() {
    const { line, column } = lineCol(view);
    pos.textContent = `${line}:${column}`;
  }
  setLineCol();
  return {
    dom,
    setMode(modeStr) {
      mode.textContent = modeStr;
    },
    setLineCol
  };
}
function lineCol(view) {
  const head = view.state.selection.main.head;
  const lineDesc = view.state.doc.lineAt(head);
  const line = lineDesc.number;
  const column = head - lineDesc.from + 1;
  return { line, column };
}
function $el(tag) {
  return document.createElement(tag);
}
function $style(el, styles) {
  Object.assign(el.style, styles);
}

// src/commands.ts
import {
  EditorSelection as EditorSelection3,
  findClusterBreak
} from "@codemirror/state";
import { matchBrackets, syntaxTree } from "@codemirror/language";
import { cursorLineStart, selectLineStart } from "@codemirror/commands";
var MODE_EFF = {
  NORMAL: modeEffect.of({
    type: 0 /* Normal */,
    minor: 2 /* Normal */
  }),
  NORMAL_GOTO: modeEffect.of({
    type: 0 /* Normal */,
    minor: 3 /* Goto */
  }),
  NORMAL_MATCH: modeEffect.of({
    type: 0 /* Normal */,
    minor: 5 /* Match */
  }),
  NORMAL_SPACE: modeEffect.of({
    type: 0 /* Normal */,
    minor: 6 /* Space */
  }),
  SELECT: modeEffect.of({
    type: 4 /* Select */,
    minor: 2 /* Normal */
  }),
  SELECT_GOTO: modeEffect.of({
    type: 4 /* Select */,
    minor: 3 /* Goto */
  }),
  SELECT_MATCH: modeEffect.of({
    type: 4 /* Select */,
    minor: 5 /* Match */
  }),
  SELECT_SPACE: modeEffect.of({
    type: 4 /* Select */,
    minor: 6 /* Space */
  }),
  INSERT: modeEffect.of({ type: 1 /* Insert */ })
};
function moveByChar(view, mode, forward) {
  const initial = view.state.selection.main;
  const select = mode.type === 4 /* Select */;
  const next = select ? selectByChar(view, mode, forward) : cursorByChar(view, mode, forward);
  if (initial.eq(next)) {
    return false;
  }
  view.dispatch({
    selection: next,
    effects: resetCount(mode),
    scrollIntoView: true
  });
  return true;
}
function withHelixSelection(view, command) {
  view.dispatch({
    selection: cmSelToInternal(view.state.selection.main, view.state.doc)
  });
  const result = command(view);
  view.dispatch({
    selection: internalSelToCM(view.state.selection.main, view.state.doc)
  });
  return result;
}
function cmSelToInternal(range, doc) {
  if (range.empty) {
    return range;
  }
  const end = nextClusterBreak(doc, range.to, false);
  const [anchor, head] = rangeForward(range) ? [range.from, end] : [end, range.from];
  return EditorSelection3.range(
    anchor,
    head,
    range.goalColumn,
    range.bidiLevel ?? void 0
  );
}
function removeText(view, { yank, edit } = {}) {
  const selection = view.state.selection.main;
  const effects = [];
  yank ??= true;
  if (yank) {
    effects.push(
      yankEffect.of([`"`, view.state.doc.slice(selection.from, selection.to)])
    );
  }
  if (edit) {
    effects.push(MODE_EFF.INSERT);
  }
  view.dispatch({
    effects,
    changes: {
      from: selection.from,
      to: selection.to,
      insert: ""
    }
  });
  if (!edit && view.state.selection.main.empty) {
    view.dispatch({
      selection: internalSelToCM(view.state.selection.main, view.state.doc)
    });
  }
}
function internalSelToCM(range, doc) {
  const end = nextClusterBreak(doc, range.to, true);
  const [anchor, head] = rangeForward(range) ? [range.from, end] : [end, range.from];
  return EditorSelection3.range(
    anchor,
    head,
    range.goalColumn,
    range.bidiLevel ?? void 0
  );
}
function cursorToLineStart(view, mode) {
  const isNormal = mode.type === 0 /* Normal */;
  withHelixSelection(
    view,
    (view2) => isNormal ? cursorLineStart(view2) : selectLineStart(view2)
  );
  view.dispatch({
    effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
  });
}
function cursorToLineEnd(view, mode) {
  const select = mode.type === 4 /* Select */;
  const selection = cmSelToInternal(view.state.selection.main, view.state.doc);
  const line = view.state.doc.lineAt(selection.head);
  if (line.length === 0) {
    return false;
  }
  const goal = nextClusterBreak(view.state.doc, line.to, false);
  const next = select ? EditorSelection3.range(selection.anchor, goal, selection.goalColumn) : EditorSelection3.cursor(goal, void 0, void 0, selection.goalColumn);
  view.dispatch({
    selection: internalSelToCM(next, view.state.doc),
    effects: select ? MODE_EFF.SELECT : MODE_EFF.NORMAL
  });
  return true;
}
function cursorByChar(view, mode, forward) {
  const doc = view.state.doc;
  const count = cmdCount(mode);
  const selection = view.state.selection.main;
  let counter = count;
  const by = count > 1 ? () => () => {
    counter--;
    return counter > 0;
  } : void 0;
  const cursor = cmSelToInternal(selection, doc).head;
  const moved = view.moveByChar(EditorSelection3.cursor(cursor), forward, by);
  return internalSelToCM(EditorSelection3.cursor(moved.head), doc);
}
function selectByChar(view, mode, forward) {
  const doc = view.state.doc;
  const count = cmdCount(mode);
  const initial = cmSelToInternal(view.state.selection.main, doc);
  let counter = count;
  const by = count > 1 ? () => () => {
    counter--;
    return counter > 0;
  } : void 0;
  const next = view.moveByChar(
    EditorSelection3.cursor(initial.head),
    forward,
    by
  );
  return internalSelToCM(EditorSelection3.range(initial.anchor, next.head), doc);
}
function selectByLine(view, mode, forward) {
  const doc = view.state.doc;
  const count = cmdCount(mode);
  const initial = cmSelToInternal(view.state.selection.main, doc);
  let selection = initial;
  for (let _i = 0; _i < count; _i++) {
    selection = view.moveVertically(
      EditorSelection3.cursor(
        selection.head,
        void 0,
        void 0,
        selection.goalColumn
      ),
      forward
    );
  }
  return internalSelToCM(
    EditorSelection3.range(initial.anchor, selection.head, selection.goalColumn),
    doc
  );
}
function cursorByLine(view, mode, forward) {
  const doc = view.state.doc;
  const count = cmdCount(mode);
  const initial = view.state.selection.main;
  const selection = cmSelToInternal(initial, doc);
  let cursor = selection.head;
  let goalColumn = selection.goalColumn;
  for (let _i = 0; _i < count; _i++) {
    const line = doc.lineAt(cursor).number;
    if (forward && line === doc.lines || !forward && line === 1) {
      break;
    }
    const next = view.moveVertically(
      EditorSelection3.cursor(cursor, void 0, void 0, goalColumn),
      forward
    );
    cursor = next.to;
    goalColumn = next.goalColumn;
  }
  return internalSelToCM(
    EditorSelection3.cursor(cursor, void 0, void 0, goalColumn),
    doc
  );
}
function moveByHalfPage(view, mode, forward) {
  const select = mode.type === 4 /* Select */;
  const next = select ? selectByHalfPage(view, forward) : cursorByHalfPage(view, forward);
  if (next.eq(view.state.selection.main)) {
    return false;
  }
  view.dispatch({
    selection: EditorSelection3.create([next]),
    scrollIntoView: true
  });
}
function cursorByHalfPage(view, forward) {
  const doc = view.state.doc;
  const selection = cmSelToInternal(view.state.selection.main, doc);
  const lineBlock = view.lineBlockAt(doc.lineAt(selection.head).from);
  const end = view.lineBlockAt(forward ? doc.length : 0);
  const height = Math.min(
    view.scrollDOM.clientHeight / 2,
    Math.abs(lineBlock.top - end.top)
  );
  if (height < 1) {
    return view.state.selection.main;
  }
  const next = view.moveVertically(
    EditorSelection3.cursor(
      selection.head,
      void 0,
      void 0,
      selection.goalColumn
    ),
    forward,
    height
  );
  return internalSelToCM(next, doc);
}
function selectByHalfPage(view, forward) {
  const doc = view.state.doc;
  const selection = cmSelToInternal(view.state.selection.main, doc);
  const lineBlock = view.lineBlockAt(doc.lineAt(selection.head).from);
  const end = view.lineBlockAt(forward ? doc.length : 0);
  const height = Math.min(
    view.scrollDOM.clientHeight / 2,
    Math.abs(lineBlock.top - end.top)
  );
  if (height < 1) {
    return view.state.selection.main;
  }
  const next = view.moveVertically(
    EditorSelection3.cursor(
      selection.head,
      void 0,
      void 0,
      selection.goalColumn
    ),
    forward,
    height
  );
  return internalSelToCM(
    EditorSelection3.range(selection.anchor, next.head, next.goalColumn),
    doc
  );
}
function moveByLine(view, mode, forward) {
  const select = mode.type === 4 /* Select */;
  const initial = view.state.selection.main;
  const next = select ? selectByLine(view, mode, forward) : cursorByLine(view, mode, forward);
  if (initial.eq(next)) {
    return false;
  }
  view.dispatch({
    selection: EditorSelection3.create([next]),
    effects: resetCount(mode),
    scrollIntoView: true
  });
  return true;
}
function moveDown(view, mode) {
  return moveByLine(view, mode, true);
}
function moveUp(view, mode) {
  return moveByLine(view, mode, false);
}
function moveLeft(view, mode) {
  moveByChar(view, mode, false);
}
function moveRight(view, mode) {
  moveByChar(view, mode, true);
}
function setFindMode(view, status, mode, metadata) {
  const effect = modeEffect.of({
    type: mode.type,
    minor: 2 /* Normal */,
    count: mode.count,
    expecting: {
      minor: status,
      callback: findText,
      metadata
    }
  });
  view.dispatch({ effects: effect });
}
function findText(view, text, {
  inclusive,
  forward
}) {
  const mode = view.state.field(modeField);
  const count = mode.type === 1 /* Insert */ ? 1 : cmdCount(mode);
  const select = mode.type === 4 /* Select */;
  const selection = cmSelToInternal(view.state.selection.main, view.state.doc);
  const doc = view.state.doc;
  const start = selection.head;
  const docString = doc.sliceString(0);
  let rawIndex = start;
  for (let _i = 0; _i < count; _i++) {
    if (!forward && rawIndex === 0) {
      rawIndex = -1;
      break;
    }
    rawIndex = forward ? docString.indexOf(text, rawIndex + 1) : docString.lastIndexOf(text, rawIndex - 1);
    if (rawIndex < 0) {
      break;
    }
  }
  const resetEffect = select ? MODE_EFF.SELECT : MODE_EFF.NORMAL;
  if (rawIndex === -1) {
    view.dispatch({
      effects: resetEffect
    });
    return;
  }
  const index = inclusive ? rawIndex : forward ? rawIndex - 1 : rawIndex + 1;
  const newSelection = select ? EditorSelection3.range(selection.anchor, index) : EditorSelection3.range(selection.head, index);
  view.dispatch({
    effects: resetEffect,
    selection: internalSelToCM(newSelection, doc)
  });
}
function moveToSibling(view, forward) {
  const tree = syntaxTree(view.state);
  const selection = view.state.selection.main;
  let stack = tree.resolveStack(selection.from, 1);
  let sibling = null;
  while (true) {
    const node = stack.node;
    if (node && node.from <= selection.from && node.to >= selection.to) {
      sibling = forward ? node?.nextSibling : node?.prevSibling;
      if (sibling) {
        break;
      }
    }
    if (stack.next) {
      stack = stack.next;
    } else {
      break;
    }
  }
  if (!sibling) {
    view.dispatch({
      selection: EditorSelection3.range(0, view.state.doc.length),
      scrollIntoView: true
    });
    return;
  }
  view.dispatch({
    selection: EditorSelection3.range(sibling.from, sibling.to),
    scrollIntoView: true
  });
}
var PAIRS = {
  "(": ["(", ")", true],
  ")": ["(", ")", false],
  "{": ["{", "}", true],
  "}": ["{", "}", false],
  "[": ["[", "]", true],
  "]": ["[", "]", false],
  "<": ["<", ">", true],
  ">": ["<", ">", false]
};
var MATCHEABLE = /* @__PURE__ */ new Set([...Object.keys(PAIRS), `"`, "'"]);
function matchBracket(view) {
  const selection = view.state.selection.main;
  const char = view.state.doc.sliceString(selection.head, selection.head + 1);
  if (!MATCHEABLE.has(char)) {
    return null;
  }
  const open = PAIRS[char]?.[2] ?? false;
  const match = matchBrackets(
    view.state,
    selection.head + (open ? 0 : 1),
    open ? 1 : -1
  );
  if (match) {
    return match.end;
  }
}
function surround(view, char, proxy) {
  const pair = PAIRS[char];
  const selection = view.state.selection.main;
  const open = pair?.[0] ?? char;
  const close = pair?.[1] ?? char;
  view.dispatch({
    changes: [
      {
        from: selection.from,
        insert: open
      },
      {
        from: selection.to,
        insert: close
      }
    ],
    effects: MODE_EFF.NORMAL
  });
  const newSelection = proxy.original.state.selection.main;
  const offset = newSelection.anchor === newSelection.from ? 1 : -1;
  const anchor = newSelection.anchor - offset;
  const head = newSelection.head + offset;
  proxy.original.dispatch({
    selection: EditorSelection3.range(anchor, head)
  });
}
function replaceWithChar(_, char, view) {
  const selection = view.state.selection.main;
  const selected = view.state.doc.sliceString(selection.from, selection.to);
  let len = 0;
  let offset = 0;
  while (true) {
    const next = findClusterBreak(selected, offset, true);
    if (next === offset) {
      break;
    }
    offset = next;
    len++;
  }
  view.dispatch({
    changes: {
      from: selection.from,
      to: selection.to,
      insert: char.repeat(len)
    },
    effects: MODE_EFF.NORMAL
  });
}
function changeCase(view, upper) {
  const selection = view.state.selection.main;
  const selected = view.state.doc.sliceString(selection.from, selection.to);
  let insert;
  if (upper == null) {
    insert = [...selected].map((char) => {
      let next = char.toUpperCase();
      return next === char ? char.toLowerCase() : next;
    }).join("");
  } else if (upper) {
    insert = selected.toUpperCase();
  } else {
    insert = selected.toLowerCase();
  }
  view.dispatch({
    changes: {
      from: selection.from,
      to: selection.to,
      insert
    }
  });
}
function paste(view, yanked, before, count, reset = true) {
  const range = view.state.selection.main;
  yanked ??= "";
  const spec = { from: range.to, insert: yanked.toString().repeat(count) };
  const change = view.state.changes(spec);
  view.dispatch(
    { changes: change },
    {
      selection: before ? { anchor: range.to, head: range.to + yanked.length } : { anchor: range.from, head: range.from + yanked.length },
      sequential: true
    },
    reset ? { effects: MODE_EFF.NORMAL } : {}
  );
}
function insertLineAndEdit(view, below) {
  let from;
  let cursor;
  if (below) {
    const line = view.state.doc.lineAt(view.state.selection.main.to);
    from = line.to;
    cursor = from + view.state.lineBreak.length;
  } else {
    const line = view.state.doc.lineAt(view.state.selection.main.from);
    from = line.from;
    cursor = from;
  }
  view.dispatch({
    changes: {
      from,
      insert: view.state.lineBreak
    },
    selection: EditorSelection3.cursor(cursor),
    effects: MODE_EFF.INSERT
  });
}
var countCommands = Object.fromEntries(
  Array.from({ length: 10 }, (_, count) => [
    String(count),
    (view, mode) => {
      const next = mode.count != null ? mode.count * 10 + count : count;
      if (next === 0) {
        return;
      }
      view.dispatch({
        effects: modeEffect.of({ ...mode, count: next })
      });
    }
  ])
);
function insertLine(view, below) {
  const mode = view.state.field(modeField);
  const count = mode.type === 1 /* Insert */ ? 1 : cmdCount(mode);
  const select = mode.type === 4 /* Select */;
  const selection = view.state.selection.main;
  const line = view.state.doc.lineAt(below ? selection.to : selection.from);
  const changes = {
    from: below ? line.to : line.from,
    insert: view.state.lineBreak.repeat(count)
  };
  const resetEffect = select ? MODE_EFF.SELECT : MODE_EFF.NORMAL;
  view.dispatch({ changes, effects: resetEffect });
}
function cmdCount(mode) {
  return mode.count ?? 1;
}
function resetCount(mode) {
  const { count: _count, ...rest } = mode;
  return modeEffect.of(rest);
}
function rangeForward(range) {
  return range.head > range.from;
}
function atomicRange(range, doc) {
  const len = range.to - range.from;
  if (len <= 1) {
    return true;
  } else if (len === 2) {
    const charCode = doc.sliceString(range.from, range.to).charCodeAt(0);
    return charCode >= 55296 && charCode <= 57343;
  } else {
    return false;
  }
}
function nextClusterBreak(doc, pos, forward) {
  if (!forward && pos === 0 || forward && pos === doc.length) {
    return pos;
  }
  const line = doc.lineAt(pos);
  if (forward && line.to === pos) {
    return pos + 1;
  } else if (!forward && line.from === pos) {
    return pos - 1;
  }
  return findClusterBreak(line.text, pos - line.from, forward) + line.from;
}

// src/search.ts
function backwardsSearch(state, query, mode, select) {
  const cursor = query.getCursor(state);
  const selection = state.selection.main;
  const count = cmdCount(mode);
  const beforeRing = new Ring(count);
  const iter = peekable(cloned(cursor));
  const beforeIter = peekingUntil(iter, (item) => item.to >= selection.from);
  for (const item of {
    [Symbol.iterator]() {
      return beforeIter;
    }
  }) {
    beforeRing.push(item);
  }
  if (beforeRing.length === count) {
    select(beforeRing.first);
    return;
  }
  const afterRing = new Ring(count - beforeRing.length);
  for (const item of {
    [Symbol.iterator]() {
      return iter;
    }
  }) {
    afterRing.push(item);
  }
  const total = afterRing.length + beforeRing.length;
  if (total === 0) {
    return {
      match: false
    };
  }
  if (total === count) {
    select(afterRing.first);
    return {
      wrapped: true
    };
  }
  const all = [...beforeRing, ...afterRing];
  const rem = count % total;
  select(all[all.length - rem - 1]);
  return {
    wrapped: true
  };
}
var Ring = class {
  items;
  head;
  length;
  maxLength;
  constructor(length) {
    this.items = Array.from({ length });
    this.maxLength = length;
    this.head = 0;
    this.length = 0;
  }
  get first() {
    return this.items[this.start];
  }
  push(item) {
    this.items[this.head] = item;
    this.head += 1;
    this.length += 1;
    this.head %= this.maxLength;
    this.length = Math.min(this.maxLength, this.length);
  }
  merge(other) {
    for (let i = 0; i < other.length; i++) {
      const item = other.items[(other.start + i) % other.maxLength];
      this.push(item);
    }
  }
  *[Symbol.iterator]() {
    for (let i = 0; i < this.length; i++) {
      yield this.items[(this.start + i) % this.maxLength];
    }
  }
  get start() {
    if (this.length < this.maxLength) {
      return 0;
    } else {
      return (this.head + 1) % this.maxLength;
    }
  }
};
function peekingUntil(iter, check) {
  return {
    next() {
      const item = iter.peek();
      if (item.done) {
        return item;
      }
      if (check(item.value)) {
        return { value: void 0, done: true };
      }
      return iter.next();
    }
  };
}
function peekable(iter) {
  let next = iter.next();
  const peekIter = {
    next() {
      const item = next;
      if (!item.done) {
        next = iter.next();
      }
      return item;
    },
    peek() {
      return next;
    }
  };
  return peekIter;
}
function cloned(iter) {
  return {
    next() {
      return { ...iter.next() };
    }
  };
}

// src/lib.ts
function startSearch(view, global) {
  const initialScroll = view.scrollSnapshot();
  const initialSelection = view.state.selection;
  let input = "";
  let query = null;
  function reset() {
    view.dispatch({
      selection: initialSelection
    });
    resetScroll(view, initialScroll);
  }
  return {
    init: view.state.field(registersField)["/"]?.toString() ?? "",
    onInput(input_) {
      if (input !== input_) {
        input = input_;
        query = searchQuery(input);
      } else {
        return;
      }
      if (!query?.valid) {
        return;
      }
      if (global) {
        return;
      }
      let match = query.getCursor(view.state, initialSelection.main.to).next();
      if (match.done) {
        match = query.getCursor(view.state).next();
      }
      if (match.done) {
        reset();
      } else {
        const selection = EditorSelection4.range(
          match.value.from,
          match.value.to
        );
        view.dispatch({
          selection,
          effects: EditorView2.scrollIntoView(selection, { y: "center" })
        });
      }
    },
    onClose(accept) {
      if (!accept) {
        reset();
        return;
      }
      if (input != null) {
        view.dispatch({
          effects: yankEffect.of(["/", input])
        });
      }
      if (global) {
        const externalCommands = view.state.facet(externalCommandsFacet);
        const query2 = input || view.state.field(registersField)["/"];
        if (query2) {
          return externalCommands.global_search?.(query2.toString());
        }
      }
    }
  };
}
function searchQuery(query) {
  return new SearchQuery({
    search: query,
    regexp: true,
    caseSensitive: /[A-Z]/.test(query)
  });
}
var searchFacet = Facet.define({
  combine(inputs) {
    const input = inputs[0];
    return input ? searchQuery(input) : void 0;
  }
});
var helixCommandBindings = {
  insert: {
    Backspace(view) {
      deleteCharBackward(view);
    },
    Delete(view) {
      deleteCharForward(view);
    },
    Enter(view) {
      insertNewlineAndIndent(view);
    },
    // we need these two due to https://github.com/codemirror/dev/issues/634
    // FIXME: stuff like Shift-<arrow> doesn't quite work with `editor.cursor-shape.insert === "block"`.
    ArrowLeft: cursorCharLeft,
    ArrowRight: cursorCharRight,
    Escape(view) {
      let selection = view.state.selection.main;
      if (selection.empty) {
        selection = internalSelToCM(selection, view.state.doc);
      }
      view.dispatch({
        effects: [
          MODE_EFF.NORMAL,
          historyEffect.of({ type: "commit", state: view.state })
        ],
        selection
      });
    }
  },
  normal: {
    // this one is special: we let it apply to all other minor modes
    Escape(view, mode_) {
      const mode = mode_;
      if (mode.type === 0 /* Normal */ && mode.minor === 2 /* Normal */ && mode.expecting == null && mode.count == null) {
        return true;
      }
      view.dispatch({
        effects: [MODE_EFF.NORMAL]
      });
    },
    ["/"](view) {
      const panel = getCommandPanel(view);
      panel.showSearchInput();
    },
    ...countCommands,
    [":"](view, mode) {
      view.dispatch({
        effects: modeEffect.of({ ...mode, count: void 0 })
      });
      getCommandPanel(view).showCommandInput();
    },
    ["y"](view) {
      const selection = view.state.selection.main;
      view.dispatch({
        effects: yankEffect.of([
          `"`,
          view.state.doc.slice(selection.from, selection.to)
        ])
      });
      getCommandPanel(view).showMessage('yanked 1 selection to register "');
    },
    ["a"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection4.cursor(selection.to)
        });
      }
    },
    ["A"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const end = view.state.doc.lineAt(selection.to).to;
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection4.cursor(end)
        });
      }
    },
    ["I"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const start = view.state.doc.lineAt(selection.from).from;
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection4.cursor(start)
        });
      }
    },
    ["c"]: {
      checkpoint: "temp",
      command(view) {
        removeText(view, { edit: true });
      }
    },
    ["d"]: {
      checkpoint: true,
      command(view) {
        removeText(view);
      }
    },
    ["Alt-c"]: {
      checkpoint: "temp",
      command(view) {
        removeText(view, { yank: false, edit: true });
      }
    },
    ["Alt-d"]: {
      checkpoint: true,
      command(view) {
        removeText(view, { yank: false });
      }
    },
    ["P"]: {
      checkpoint: true,
      command(view, mode) {
        const yanked = view.state.field(registersField);
        paste(view, yanked[`"`], false, cmdCount(mode));
      }
    },
    ["p"]: {
      checkpoint: true,
      command(view, mode) {
        const yanked = view.state.field(registersField);
        paste(view, yanked[`"`], true, cmdCount(mode));
      }
    },
    ["R"]: {
      checkpoint: true,
      command(view, mode) {
        const contents = view.state.field(registersField)[`"`] ?? "";
        const count = cmdCount(mode);
        const replacement = count === 1 ? contents : contents.toString().repeat(count);
        const tr = view.state.replaceSelection(replacement);
        tr.effects = MODE_EFF.NORMAL;
        view.dispatch(tr);
      }
    },
    ["r"]: {
      checkpoint: true,
      command(view, mode) {
        view.dispatch({
          effects: modeEffect.of({
            ...mode,
            expecting: {
              minor: "r",
              callback: replaceWithChar,
              metadata: view
            }
          })
        });
      }
    },
    ["w"](view, mode) {
      if (mode.type === 0 /* Normal */) {
        const current = view.state.selection.main;
        view.dispatch({
          selection: EditorSelection4.single(Math.max(current.from, current.to))
        });
      }
      for (let _i = 0; _i < cmdCount(mode); _i++) {
        selectGroupRight(view);
      }
    },
    ["b"](view, mode) {
      if (mode.type === 0 /* Normal */) {
        const current = view.state.selection.main;
        view.dispatch({
          selection: EditorSelection4.single(Math.min(current.from, current.to))
        });
      }
      for (let _i = 0; _i < cmdCount(mode); _i++) {
        selectGroupLeft(view);
      }
    },
    ["v"](view, mode) {
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.SELECT : MODE_EFF.NORMAL
      });
    },
    ["g"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL_GOTO : MODE_EFF.SELECT_GOTO
      });
    },
    ["Space"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL_SPACE : MODE_EFF.SELECT_SPACE
      });
    },
    ["m"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL_MATCH : MODE_EFF.SELECT_MATCH
      });
    },
    ["i"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const cursor = Math.min(selection.from, selection.to);
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection4.cursor(cursor)
        });
      }
    },
    ["h"](view, mode) {
      return moveLeft(view, mode);
    },
    ["j"]: moveDown,
    ["ArrowDown"]: "j",
    ["ArrowUp"]: "k",
    ["k"]: moveUp,
    ["ArrowRight"]: "l",
    ["ArrowLeft"]: "h",
    ["l"](view, mode) {
      return moveRight(view, mode);
    },
    ["%"](view) {
      return selectAll(view);
    },
    ["o"]: {
      checkpoint: "temp",
      command(view) {
        insertLineAndEdit(view, true);
      }
    },
    ["O"]: {
      checkpoint: "temp",
      command(view) {
        insertLineAndEdit(view, false);
      }
    },
    ["f"](view, mode) {
      setFindMode(view, "f", mode, {
        inclusive: true,
        forward: true
      });
    },
    ["F"](view, mode) {
      setFindMode(view, "F", mode, {
        inclusive: true,
        forward: false
      });
    },
    ["t"](view, mode) {
      setFindMode(view, "t", mode, {
        inclusive: false,
        forward: true
      });
    },
    ["T"](view, mode) {
      setFindMode(view, "T", mode, {
        inclusive: false,
        forward: false
      });
    },
    ["u"](view, mode) {
      const { checkpoints, cursor } = view.state.field(historyField);
      const nextCursor = cursor + cmdCount(mode);
      const state = checkpoints[nextCursor];
      if (!state) {
        return true;
      }
      view.dispatch({
        changes: {
          from: 0,
          to: view.state.doc.length,
          insert: state.doc
        },
        selection: state.selection,
        effects: [
          historyEffect.of({
            type: "move",
            offset: cmdCount(mode),
            head: view.state
          }),
          resetCount(mode)
        ],
        scrollIntoView: true
      });
    },
    ["U"](view, mode) {
      const { checkpoints, cursor, head } = view.state.field(historyField);
      const nextCursor = cursor - cmdCount(mode);
      const state = nextCursor === -1 ? head : checkpoints[nextCursor];
      if (!state) {
        return true;
      }
      view.dispatch({
        changes: {
          from: 0,
          to: view.state.doc.length,
          insert: state.doc
        },
        selection: state.selection,
        effects: [
          historyEffect.of({
            type: "move",
            offset: -cmdCount(mode)
          }),
          resetCount(mode)
        ],
        scrollIntoView: true
      });
    },
    ["x"](view, mode) {
      const initial = view.state.selection.main;
      const startLine = view.state.doc.lineAt(initial.from);
      let endLine = view.state.doc.lineAt(initial.to);
      if (!initial.empty && initial.to === endLine.from) {
        endLine = view.state.doc.line(endLine.number - 1);
      }
      const ideal = EditorSelection4.range(
        startLine.from,
        Math.min(
          view.state.doc.length,
          endLine.to + view.state.lineBreak.length
        )
      );
      let nextSel;
      const perfectLineSelection = ideal.from === initial.from && ideal.to === initial.to;
      if (perfectLineSelection || mode.count) {
        const nextLineNumber = Math.min(
          endLine.number + cmdCount(mode),
          view.state.doc.lines
        );
        const nextLine = view.state.doc.line(nextLineNumber);
        nextSel = EditorSelection4.range(
          startLine.from,
          Math.min(
            view.state.doc.length,
            nextLine.to + view.state.lineBreak.length
          )
        );
      } else {
        nextSel = ideal;
      }
      view.dispatch({
        selection: nextSel,
        effects: resetCount(mode)
      });
      return true;
    },
    ["n"](view, mode) {
      const query = view.state.facet(searchFacet);
      if (!query?.valid) {
        if (query) {
          showSearchError(view, query);
        }
        view.dispatch({
          effects: resetCount(mode)
        });
        return true;
      }
      let cursor = query.getCursor(view.state, view.state.selection.main.to);
      let match;
      let found = false;
      let wrapped = false;
      for (let _i = 0; _i < cmdCount(mode); _i++) {
        match = cursor.next();
        found ||= !match.done;
        if (match.done) {
          cursor = query.getCursor(view.state);
          wrapped = true;
          match = cursor.next();
          found ||= !match.done;
        }
        if (!found) {
          getCommandPanel(view).showError("No more matches");
          return;
        }
      }
      view.dispatch({
        selection: EditorSelection4.range(match.value.from, match.value.to),
        scrollIntoView: true
      });
      if (wrapped) {
        getCommandPanel(view).showMessage("Wrapped around document");
      }
    },
    ["N"](view, mode) {
      const query = view.state.facet(searchFacet);
      if (!query?.valid) {
        if (query) {
          showSearchError(view, query);
        }
        view.dispatch({
          effects: resetCount(mode)
        });
        return true;
      }
      const result = backwardsSearch(view.state, query, mode, (match) => {
        const selection = EditorSelection4.range(match.from, match.to);
        view.dispatch({
          selection,
          effects: EditorView2.scrollIntoView(selection, { y: "center" })
        });
      });
      if (result?.wrapped) {
        getCommandPanel(view).showMessage("Wrapped around document");
      } else if (result?.match === false) {
        getCommandPanel(view).showError("No more matches");
      }
      view.dispatch({
        effects: resetCount(mode)
      });
    },
    ["Ctrl-d"](view, mode) {
      moveByHalfPage(view, mode, true);
    },
    ["PageDown"]: "Ctrl-d",
    ["PageUp"]: "Ctrl-u",
    ["Ctrl-u"](view, mode) {
      moveByHalfPage(view, mode, false);
    },
    [";"](view) {
      withHelixSelection(view, () => {
        if (view.state.selection.main.empty) {
          return false;
        }
        view.dispatch({
          selection: EditorSelection4.cursor(view.state.selection.main.head),
          scrollIntoView: true
        });
        return true;
      });
    },
    ["Alt-;"](view) {
      const selection = view.state.selection.main;
      if (atomicRange(selection, view.state.doc)) {
        return;
      }
      view.dispatch({
        selection: EditorSelection4.range(selection.head, selection.anchor),
        scrollIntoView: true
      });
    },
    ["Alt-:"](view) {
      const selection = view.state.selection.main;
      if (atomicRange(selection, view.state.doc)) {
        return;
      }
      view.dispatch({
        selection: EditorSelection4.range(selection.from, selection.to)
      });
    },
    ["Alt-ArrowUp"](view) {
      expandSyntaxHistory(
        view.state,
        (start, dispatch) => {
          view.dispatch(start);
          selectParentSyntax({
            state: view.state,
            dispatch
          });
        },
        (spec) => view.dispatch(spec)
      );
    },
    ["Alt-o"]: "Alt-ArrowUp",
    ["Alt-i"]: "Alt-ArrowDown",
    ["Alt-ArrowDown"](view) {
      const result = undoSyntaxHistory(view.state);
      if (result) {
        view.dispatch(result);
      }
    },
    ["Alt-ArrowRight"](view) {
      moveToSibling(view, true);
    },
    ["Alt-n"]: "Alt-ArrowRight",
    ["Alt-ArrowLeft"](view) {
      moveToSibling(view, false);
    },
    ["Alt-p"]: "Alt-ArrowLeft",
    ["Ctrl-c"]: {
      checkpoint: true,
      command(view) {
        toggleComment(view);
        view.original.dispatch({
          effects: MODE_EFF.NORMAL
        });
      }
    },
    [">"]: {
      checkpoint: true,
      command(view) {
        return indentMore(view);
      }
    },
    ["<"]: {
      checkpoint: true,
      command(view) {
        return indentLess(view);
      }
    },
    ["`"]: {
      checkpoint: true,
      command(view) {
        changeCase(view, false);
      }
    },
    ["Alt-`"]: {
      checkpoint: true,
      command(view) {
        changeCase(view, true);
      }
    },
    ["~"]: {
      checkpoint: true,
      command(view) {
        changeCase(view);
      }
    },
    ["*"](view) {
      const selection = view.state.selection.main;
      const selected = view.state.doc.slice(selection.from, selection.to).toString();
      const yanked = escapeRegex(selected);
      view.dispatch({
        effects: yankEffect.of(["/", yanked])
      });
      getCommandPanel(view).showMessage(`register '/' set to '${yanked}'`);
    },
    ["_"](view) {
      const selection = view.state.selection.main;
      const selected = view.state.doc.slice(selection.from, selection.to).toString();
      const trimmed = selected.trim();
      if (trimmed === selected) {
        return;
      }
      const startOffset = selected.indexOf(trimmed);
      const endOffset = selected.length - trimmed.length - startOffset;
      const anchor = selection.anchor === selection.from ? selection.anchor + startOffset : selection.anchor - endOffset;
      const head = selection.head === selection.to ? selection.head - endOffset : selection.head + startOffset;
      view.dispatch({
        selection: EditorSelection4.range(anchor, head)
      });
    },
    ["Home"]: cursorToLineStart,
    ["End"]: cursorToLineEnd,
    ["J"]: {
      checkpoint: true,
      command(view) {
        const selection = view.state.selection.main;
        const { doc } = view.state;
        const startLine = doc.lineAt(selection.from);
        if (startLine.number >= doc.lines) {
          return;
        }
        let endLine = doc.lineAt(selection.to);
        const sameLine = endLine.number === startLine.number;
        if (sameLine) {
          endLine = doc.line(startLine.number + 1);
        }
        let content = "";
        let removed = 0;
        for (let lineNo = startLine.number; lineNo <= endLine.number; lineNo++) {
          let lineContent = startLine.text;
          if (lineNo > startLine.number) {
            const lineText = doc.line(lineNo).text;
            lineContent = lineText.trimStart();
            let trimmed = lineText.length - lineContent.length;
            if (!sameLine && lineNo === endLine.number && selection.to - endLine.from < trimmed) {
              trimmed = selection.to - endLine.from;
            }
            removed += trimmed;
          }
          content += lineContent;
          content += lineNo === endLine.number ? "" : " ";
        }
        const newTo = sameLine ? selection.to : selection.to - removed - (endLine.number - startLine.number) * (view.state.lineBreak.length - 1);
        view.dispatch({
          changes: {
            from: startLine.from,
            to: endLine.to,
            insert: content
          },
          selection: selection.anchor > selection.head ? EditorSelection4.range(newTo, selection.from) : EditorSelection4.range(selection.from, newTo)
        });
      }
    },
    ["["](view, mode) {
      view.dispatch({
        effects: modeEffect.of({
          type: mode.type,
          minor: 7 /* LeftBracket */
        })
      });
    },
    ["]"](view, mode) {
      view.dispatch({
        effects: modeEffect.of({
          type: mode.type,
          minor: 8 /* RightBracket */
        })
      });
    }
  },
  goto: {
    ["g"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      withHelixSelection(
        view,
        () => isNormal ? cursorDocStart(view) : selectDocStart(view)
      );
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["e"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      withHelixSelection(
        view,
        () => isNormal ? cursorDocEnd(view) : selectDocEnd(view)
      );
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["h"]: cursorToLineStart,
    ["j"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      moveDown(view, mode);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["k"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      moveUp(view, mode);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["l"]: cursorToLineEnd,
    ["n"](view, mode) {
      const externalCommandDefs = view.state.facet(externalCommandsFacet);
      const result = externalCommandDefs[":buffer-next"]?.();
      getCommandPanel(view).showMessage(result);
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["p"](view, mode) {
      const externalCommandDefs = view.state.facet(externalCommandsFacet);
      const result = externalCommandDefs[":buffer-previous"]?.();
      getCommandPanel(view).showMessage(result);
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    }
  },
  match: {
    ["s"]: {
      checkpoint: true,
      command(view, mode) {
        view.dispatch({
          effects: modeEffect.of({
            type: mode.type,
            minor: 5 /* Match */,
            expecting: {
              minor: "s",
              callback: surround,
              metadata: view
            }
          })
        });
      }
    },
    ["m"](view, mode) {
      const selection = matchBracket(view) ?? void 0;
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        selection: selection?.from == null ? void 0 : isNormal ? EditorSelection4.cursor(selection.from) : EditorSelection4.range(
          view.state.selection.main.anchor,
          selection.from
        ),
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT,
        scrollIntoView: true
      });
    }
  },
  space: {
    ["y"](view) {
      const selection = view.state.selection.main;
      navigator.clipboard.writeText(
        view.state.doc.slice(selection.from, selection.to).toString()
      );
      getCommandPanel(view).showMessage("yanked 1 selection to register +");
      view.dispatch({
        effects: MODE_EFF.NORMAL
      });
    },
    ["p"]: {
      checkpoint: true,
      command(view) {
        view.dispatch({ effects: MODE_EFF.NORMAL });
        navigator.clipboard.readText().then((yanked) => paste(view, yanked, true, 1, false));
      }
    },
    ["P"]: {
      checkpoint: true,
      command(view) {
        view.dispatch({ effects: MODE_EFF.NORMAL });
        navigator.clipboard.readText().then((yanked) => paste(view, yanked, false, 1, false));
      }
    },
    ["R"]: {
      checkpoint: true,
      command(view) {
        view.dispatch({ effects: MODE_EFF.NORMAL });
        navigator.clipboard.readText().then((yanked) => {
          const tr = view.state.replaceSelection(yanked);
          view.dispatch(tr);
        });
      }
    },
    ["f"](view, mode) {
      const externalCommandDefs = view.state.facet(externalCommandsFacet);
      const result = externalCommandDefs.file_picker?.();
      getCommandPanel(view).showMessage(result);
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["b"](view, mode) {
      const externalCommandsDef = view.state.facet(externalCommandsFacet);
      const result = externalCommandsDef.buffer_picker?.();
      getCommandPanel(view).showMessage(result);
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["/"](view, mode) {
      const enabled = view.state.facet(externalCommandsFacet).global_search != null;
      if (enabled) {
        getCommandPanel(view).showSearchInput(true);
      }
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    }
  },
  leftBracket: {
    ...countCommands,
    ["Space"]: {
      checkpoint: true,
      command(view) {
        insertLine(view, false);
      }
    }
  },
  rightBracket: {
    ...countCommands,
    ["Space"]: {
      checkpoint: true,
      command(view) {
        insertLine(view, true);
      }
    }
  }
};
function toCodemirrorKeymap(keybindings) {
  const allKeys = [
    ...new Set(
      Object.values(keybindings).flat().flatMap((binding) => Object.keys(binding))
    )
  ];
  function apply(def, view, mode) {
    if (typeof def === "function") {
      return def(view, mode);
    } else {
      const temp = def.checkpoint === "temp";
      return def.command(
        {
          original: view,
          dispatch(...args) {
            view.dispatch(commitToHistory(view, temp), ...args);
          },
          get state() {
            return view.state;
          }
        },
        mode
      );
    }
  }
  function getExplicitCommand(key, bindings) {
    while (true) {
      const binding = bindings[key];
      if (typeof binding === "string") {
        key = binding;
        continue;
      }
      return binding;
    }
  }
  const codemirrorKeybindings = [];
  for (const key of allKeys) {
    const insertCommand = getExplicitCommand(key, keybindings.insert);
    const normalCommand = getExplicitCommand(key, keybindings.normal);
    const gotoCommand = getExplicitCommand(key, keybindings.goto);
    const matchCommand = getExplicitCommand(key, keybindings.match);
    const spaceCommand = getExplicitCommand(key, keybindings.space);
    const leftBracketCommand = getExplicitCommand(
      key,
      keybindings.leftBracket
    );
    const rightBracketCommand = getExplicitCommand(
      key,
      keybindings.rightBracket
    );
    const esc = key === "Escape";
    const isChar = key.length === 1 || key === "Space";
    const command = (view) => {
      const mode = view.state.field(modeField);
      if (mode.type === 1 /* Insert */) {
        if (insertCommand) {
          return insertCommand(view, void 0) ?? true;
        } else {
          return false;
        }
      }
      if (mode.expecting && isChar) {
        return false;
      }
      let result;
      if (esc || mode.minor === 2 /* Normal */ && normalCommand) {
        result = apply(normalCommand, view, mode);
      } else if (mode.minor === 3 /* Goto */ && gotoCommand) {
        result = apply(gotoCommand, view, mode);
      } else if (mode.minor === 5 /* Match */ && matchCommand) {
        result = apply(matchCommand, view, mode);
      } else if (mode.minor === 6 /* Space */ && spaceCommand) {
        result = apply(spaceCommand, view, mode);
      } else if (mode.minor === 7 /* LeftBracket */ && leftBracketCommand) {
        result = apply(leftBracketCommand, view, mode);
      } else if (mode.minor === 8 /* RightBracket */ && rightBracketCommand) {
        result = apply(rightBracketCommand, view, mode);
      } else {
        return false;
      }
      return result ?? true;
    };
    codemirrorKeybindings.push({
      key,
      run: command
    });
  }
  return codemirrorKeybindings;
}
var EndLineCursor = class extends WidgetType {
  toDOM() {
    const span = document.createElement("span");
    span.classList.add("cm-hx-cursor");
    span.classList.add("cm-hx-cursor-endline");
    span.textContent = " ";
    return span;
  }
};
var cursorMark = Decoration.mark({ class: "cm-hx-cursor" });
var endlineCursorWidget = Decoration.widget({
  widget: new EndLineCursor(),
  side: 1
});
function drawCursorMark(selection, doc) {
  const headSel = internalSelToCM(
    EditorSelection4.cursor(cmSelToInternal(selection.main, doc).head),
    doc
  );
  const line = doc.lineAt(headSel.head);
  if (headSel.from === doc.length || line.to === headSel.from) {
    return Decoration.set(
      endlineCursorWidget.range(headSel.from, headSel.from)
    );
  } else {
    return Decoration.set(cursorMark.range(headSel.head, headSel.anchor));
  }
}
function letThrough(tr) {
  return tr;
}
var selectByClickFilter = EditorState3.transactionFilter.from(
  modeField,
  (mode) => mode.type === 1 /* Insert */ ? letThrough : (tr) => {
    const userEvent = tr.annotation(Transaction2.userEvent);
    if (userEvent !== "select.pointer") {
      return tr;
    }
    const selection = tr.newSelection.main;
    if (!selection.empty) {
      return tr;
    }
    return [
      tr,
      {
        selection: internalSelToCM(selection, tr.newDoc)
      }
    ];
  }
);
var unhandledCommandsFilter = EditorState3.transactionFilter.from(
  modeField,
  (mode) => mode.type === 1 /* Insert */ ? letThrough : (tr) => {
    const userEvent = tr.annotation(Transaction2.userEvent);
    if (userEvent == null) {
      return tr;
    }
    if (!userEvent.startsWith("input")) {
      return tr;
    }
    if (!userEvent.startsWith("input.type")) {
      return tr;
    }
    if (mode.minor !== 2 /* Normal */) {
      return {
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      };
    }
    return [];
  }
);
var inputHandler = EditorView2.inputHandler.from(
  modeField,
  (mode) => (view, _from, _to, text) => {
    if (mode.type === 1 /* Insert */) {
      return false;
    }
    if (mode.expecting) {
      mode.expecting.callback(view, text, mode.expecting.metadata);
      return true;
    }
    return false;
  }
);
var updateListener = EditorView2.updateListener.of((viewUpdate) => {
  const { state, startState } = viewUpdate;
  const panel = getPanel(viewUpdate.view, statusPanel);
  const mode = state.field(modeField);
  const startMode = startState.field(modeField);
  if (mode !== startMode) {
    const startExternalMode = toExternalMode(startMode);
    const externalMode = toExternalMode(mode);
    if (startExternalMode !== externalMode) {
      panel.setMode(externalMode);
    }
  }
  panel.setLineCol();
});
var helixKeymap = keymap.of(toCodemirrorKeymap(helixCommandBindings));
var externalCommandsFacet = Facet.define({
  combine(values) {
    const handlers = [...values];
    handlers.reverse();
    if (true) {
      const merged = values.reduce((acc, defs) => {
        for (const key of Object.keys(defs)) {
          if (acc[key] == null) {
            acc[key] = 1;
          } else {
            acc[key]++;
          }
        }
        return acc;
      }, {});
      const multiple = Object.entries(merged).flatMap(
        ([key, count]) => count > 1 ? [key] : []
      );
      if (multiple.length > 0) {
        console.warn(
          `Multiple definitions found for external commands: ${multiple.join(
            ", "
          )}`
        );
      }
    }
    return handlers.reduce((acc, def) => {
      return { ...acc, ...def };
    }, {});
  }
});
function snapshot(state, global = false) {
  return {
    registers: state.field(registersField),
    ...global ? {} : {
      history: state.field(historyField)
    }
  };
}
function globalStateSync(state) {
  return [
    {
      effects: yankEffect.of({ reset: state.field(registersField) })
    }
  ];
}
var commands = Facet.define({
  combine(commands2) {
    return commands2.flat();
  }
});
var resetMode = MODE_EFF.NORMAL;
function helix(options = {}) {
  const cursorShape = options?.config?.["editor.cursor-shape.insert"] ?? "block";
  const registerState = options.globalInit ?? options.init;
  const initialRegisters = registerState instanceof EditorState3 ? registerState.field(registersField) : registerState ? registerState.registers : void 0;
  const initialHistory = options.init instanceof EditorState3 ? options.init.field(historyField) : options.init ? options.init.history : void 0;
  return [
    EditorView2.theme({
      ".cm-hx-block-cursor .cm-cursor": {
        display: "none !important"
      },
      ".cm-hx-block-cursor .cm-hx-cursor": {
        background: "#ccc"
      }
    }),
    panelStyles,
    drawSelection({
      cursorBlinkRate: 0,
      drawRangeCursor: cursorShape === "bar"
    }),
    helixKeymap,
    modeField,
    initialHistory ? historyField.init(() => initialHistory) : historyField,
    initialRegisters ? registersField.init(() => initialRegisters) : registersField,
    searchFacet.from(registersField, (registers) => registers["/"]?.toString()),
    unhandledCommandsFilter,
    selectByClickFilter,
    inputHandler,
    EditorState3.transactionFilter.from(
      syntaxHistoryField,
      ({ selections }) => selections.length === 0 ? letThrough : (tr) => {
        for (const effect of tr.effects) {
          if (effect.is(syntaxHistoryEffect)) {
            return tr;
          }
        }
        return [tr, { effects: syntaxHistoryEffect.of({ type: "reset" }) }];
      }
    ),
    EditorView2.decorations.compute(["selection", "doc", modeField], (state) => {
      if (cursorShape === "bar" && state.field(modeField).type === 1 /* Insert */) {
        return Decoration.set([]);
      }
      return drawCursorMark(state.selection, state.doc);
    }),
    updateListener,
    showPanel.of(statusPanel),
    showPanel.of(commandPanel),
    syntaxHistoryField,
    ViewPlugin.define((view) => {
      view.scrollDOM.classList.add("cm-hx-block-cursor");
      if (view.state.doc.length !== 0) {
        setTimeout(() => {
          view.dispatch({
            selection: EditorSelection4.range(1, 0)
          });
        });
      }
      return {
        update(update) {
          const mode = update.state.field(modeField);
          const startMode = update.startState.field(modeField);
          const panel = getCommandPanel(view);
          if (panel.hasMessage() && update.docChanged || update.selectionSet) {
            panel.clearMessage();
          }
          const modeChanged = !sameMode(mode, startMode);
          if (modeChanged || !sameModeState(mode, startMode)) {
            panel.showMinor(mode);
          }
          if (modeChanged && cursorShape === "bar") {
            view.scrollDOM.classList.toggle(
              "cm-hx-block-cursor",
              mode.type !== 1 /* Insert */
            );
          }
        }
      };
    }),
    commands.of([
      {
        name: "goto",
        aliases: ["g"],
        help: "Goto line number",
        handler(view, args) {
          if (args.length === 0) {
            return { message: "Line number required", error: true };
          }
          const lineNo = Number(args[0]);
          if (!Number.isFinite(lineNo) || lineNo <= 0) {
            return { message: "Invalid line number", error: true };
          }
          const effectiveLine = Math.min(lineNo, view.state.doc.lines);
          const line = view.state.doc.line(effectiveLine);
          view.dispatch({
            selection: EditorSelection4.cursor(line.from),
            scrollIntoView: true
          });
        }
      },
      {
        name: "clipboard-yank",
        help: "Yank main selection into system clipboard",
        handler(view) {
          const selection = view.state.selection.main;
          navigator.clipboard.writeText(
            view.state.doc.slice(selection.from, selection.to).toString()
          );
          return { message: "Yanked main selection to + register" };
        }
      }
    ]),
    commands.compute([externalCommandsFacet], (state) => {
      const externalCommands = state.facet(externalCommandsFacet);
      const hardcodedCommands = [
        [":buffer-next", "Goto next buffer", ["bn", "bnext"]],
        [":buffer-previous", "Goto previous buffer", ["bp", "bprev"]],
        [":buffer-close", "Close the current buffer", ["bc", "bclose"]]
      ];
      return hardcodedCommands.filter(([name]) => !!externalCommands[name]).map(([name, help, aliases]) => ({
        name: name.slice(1),
        aliases,
        help,
        handler(view) {
          const defs = view.state.facet(externalCommandsFacet);
          return defs[name]?.();
        }
      }));
    })
  ];
}
function commandPanel(view) {
  return new CommandPanel(
    view,
    commands,
    (global) => startSearch(view, global)
  );
}
function getCommandPanel(view) {
  return getPanel(view, commandPanel);
}
function toExternalMode(mode) {
  switch (mode.type) {
    case 0 /* Normal */:
      return "NOR";
    case 4 /* Select */:
      return "SEL";
    case 1 /* Insert */:
      return "INS";
  }
}
function commitToHistory(view, temp = false) {
  return {
    effects: historyEffect.of({
      type: "add",
      state: view.state,
      temp
    })
  };
}
function showSearchError(view, query) {
  let message = "";
  try {
    query.getCursor(view.state);
  } catch (error) {
    message = error?.message;
  }
  getCommandPanel(view).showError(
    `Invalid regex /${query.search}/: ${message}`
  );
}
function resetScroll(view, effect) {
  requestAnimationFrame(
    () => requestAnimationFrame(() => {
      view.dispatch({ effects: effect });
    })
  );
}
function escapeRegex(text) {
  return text.replace(/[|\\{}()[\]^$+*?.]/g, "\\$&").replace(/-/g, "\\x2d");
}
export {
  commands,
  externalCommandsFacet as externalCommands,
  globalStateSync,
  helix,
  resetMode,
  snapshot
};
